//********************************************************************
//*** ResourcePool/SQL/RTSQLSelectRow.hh
//*** Copyright (c) 2003-2008 by Markus Winand <mws@fatalmind.com>
//*** $Id: RTSQLSelectRow.hh,v 1.20 2008/05/25 10:23:46 mws Exp $
//********************************************************************
/*
This file is part of ResourcePool.

ResourcePool is free software; you can redistribute it
and/or modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

ResourcePool is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied warranty
of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ResourcePool; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307  USA */

#ifndef RESOURCEPOOL_SQL_RTSQLSELECTROW_HH
#define RESOURCEPOOL_SQL_RTSQLSELECTROW_HH

#include "mm/gc_ptr.hh"
#include "ResourcePool/SQL/RTSQLCommand.hh"
#include "ResourcePool/SQL/Resource.hh"
#include "ResourcePool/SQL/BindInterface.hh"
#include "ResourcePool/SQL/BindOutInterface.hh"
#include "ResourcePool/SQL/RTCommandParameters.hh"
#include "ResourcePool/SQL/SQLFetcher.hh"

namespace fatalmind {
namespace SQL {

template<class TM> class RTSQLSelectRow_Impl;


/**
 * A command to issue a select statement against
 * an SQL database and fetch only a single row.
 * 
 * If the number of rows returned by the statement 
 * is not equal to 1 an exception will be thrown.
 *
 * The exception might be thrown AFTER the SQLFetcher's
 * process() method was called.
 *
 * Please refer to the <a href="selectrow_8cc-example.html">selectrow.cc example</a> to see it in action.
 */
template<class TM>
class RTSQLSelectRow
: public RTSQLCommand<TM>
, public BindInterface
, public BindOutInterface
{
	public:
		RTSQLSelectRow(const ResourcePool<ResourceType<TM> >& p, const std::string& sql, SQL::SQLFetcher& fetcher = SQL::NopFetcher);
		virtual ~RTSQLSelectRow();
		
		template<typename T>
		void bind(const unsigned int pos, const T& val, bool nul = false);

		void clearParameters();

		template<typename T>
		void bindout(const unsigned int pos, T& val, bool& isnull);

		template<typename T>
		void bindout(const unsigned int pos, T& val, BindOutInterface::BindOutNotNullConstraint = NOTNULL);

//		bool isnull(const unsigned int pos) const;

		void execute(typename ResourceType<TM>::resource_t& hand);

		virtual void outputoperator(std::ostream&) const;
		virtual const Clone* getPlainSpecializedCommand() const;
	protected:
		virtual Clone* DoClone() const {
			return new RTSQLSelectRow<TM>(*this);
		}
	private:
		gc_ptr<RTSQLSelectRow_Impl<TM> > _realSQLExecute;	

		typedef typename SQLFactory<TM>::databaseType_t databaseType_t;
};


} // namespace SQL
} // namespace fatalmind
#endif
