//********************************************************************
//*** ResourcePool/oracle/Transaction.cc
//*** Copyright (c) 2003-2008 by Markus Winand <mws@fatalmind.com>
//*** $Id: Transaction.cc,v 1.29 2008/04/14 13:40:22 mws Exp $
//********************************************************************
/*
This file is part of ResourcePool.

ResourcePool is free software; you can redistribute it
and/or modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2 of 
the License, or (at your option) any later version.

ResourcePool is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied warranty
of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ResourcePool; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307  USA */

#include <iostream>
#include <memory>

#ifndef RESOURCEPOOL_ORACLE_TRANSACTION_HH
#include "Transaction.hh"
#endif

#ifndef RESOURCEPOOL_ORACLE_RESOURCE_HH
#include "Resource.hh"
#endif

#ifndef RESOURCEPOOL_ORACLE_ORACLEEXCEPTIONFACTORY_HH
#include "OracleExceptionFactory.hh"
#endif

#ifndef RESOURCEPOOL_ORACLE_RESOURCEPOOL_HH
#include "ResourcePool.hh"
#endif

namespace fatalmind {
namespace oracle {

template<class TM, template<class> class SLMDC>
Transaction<TM, SLMDC>::Transaction(const Command<ResourceType<TM, SLMDC> >& a_cmd)
: cmd(a_cmd.template clone<Command<ResourceType<TM, SLMDC> > >())
{
}

template<class TM, template<class> class SLMDC>
Transaction<TM, SLMDC>::Transaction(const ResourcePool<ResourceType<TM> >& p, const Command<ResourceType<TM, SLMDC> >& a_cmd)
: cmd(a_cmd.template clone<Command<ResourceType<TM, SLMDC> > >())
{
}

template<class TM, template<class> class SLMDC>
Transaction<TM, SLMDC>::Transaction()
: cmd()
{
}

// TODO, is this implementation correct, shouldn't it clone the rho's cmd?
template<class TM, template<class> class SLMDC>
Transaction<TM, SLMDC>::Transaction(const Transaction<TM, SLMDC>& rho)
: cmd(rho.template clone<CommandInterface<TM, SLMDC> >())
{
}

template<class TM, template<class> class SLMDC>
Transaction<TM, SLMDC>::~Transaction() {
}

template<class TM, template<class> class SLMDC>
Clone* Transaction<TM, SLMDC>::DoClone() const {
	return new Transaction(*this);
}

template<class TM, template<class> class SLMDC>
void Transaction<TM, SLMDC>::commit(oracleHandlesWithCache<TM>& hand) {
	sword ret;
	if ((ret = OCITransCommit(hand.getSvcCtx()
		, hand.getError()
		, OCI_DEFAULT
	)) != OCI_SUCCESS) {
		OracleExceptionFactory::throwit("OCITransCommit", hand, ret, "commit");
	}
}

template<class TM, template<class> class SLMDC>
void Transaction<TM, SLMDC>::rollback(oracleHandlesWithCache<TM>& hand) {
	sword ret;
	if ((ret = OCITransRollback(hand.getSvcCtx()
		, hand.getError()
		, OCI_DEFAULT
	)) != OCI_SUCCESS) {
		OracleExceptionFactory::throwit("OCITransRollback", hand, ret, "rollback");
	}
}

template<class TM, template<class> class SLMDC>
void Transaction<TM, SLMDC>::execute(oracleResource<TM>& hand) {
	StatementCache sthcache;
	_execute(hand, sthcache);
}

template<class TM, template<class> class SLMDC>
void Transaction<TM, SLMDC>::_execute(oracleResource<TM>& hand, StatementCache& sthcache) {
	if (cmd.get()) {
		try {
			FinallyTxDepth<TM> txCounter(hand);
			cmd->execute(hand);
			commit(hand);
		} catch (...) {
			rollback(hand);
			throw;
		}
	}
}

template class Transaction<DefaultThreadedModel, oracle::SelectListMetaDataCache >;

} // namespace oracle
} // namespace fatalmind
