#include <iostream>
#include <iomanip>
#include <sstream>
#include "ResourcePool/SQL/ResourcePool.hh"
                                          /* define the resource type to use */
typedef fatalmind::ResourcePool<fatalmind::ResourceType::SQL> RP;

int parseString(char*);

class PrintFetcher: public fatalmind::SQL::SQLFetcher {
	public:
		PrintFetcher(int c) 
		: columns(c)
		, row(columns)
		{
			/* calls the string's default c-tor n-times */
			row.resize(columns);
			width.resize(columns);
		};

		void bindout(RP::Command::SQLSelect& cmd) {
			for (int i = 0; i < columns; ++i) {
				cmd.bindout(i, row[i].first, row[i].second);
			}
		}

		bool process(int rownum) {
			std::cout << "|  ";
			for (int i = 0; i < columns; ++i) {
				if (! row[i].second) {
					int len = row[i].first.length();
					width[i] = std::max(len, width[i]);
					std::cout << std::setw(width[i]) << row[i].first << "  |  ";
				} else {
					std::cout << "|  ";
				}
			}
			std::cout << std::endl;
		}

		void done(bool more, int fetched) {
			std::cout << std::endl << (more ? "": "all ") << fetched << " row" << (fetched != 1 ? "s":"") << " fetched." <<std::endl << std::endl;
		}
	private:
		int columns;
		typedef std::pair<std::string, bool> column;
		std::vector<column> row;
		std::vector<int> width;
};

int main(int argc, char* argv[]) {
	try {
		if (argc < 4) {                   /* just some minimalistic checking */
			throw fatalmind::Exception("arguments missing: driver:logon no-columns SQL");
		}
		RP::Factory f(argv[1]);           /* pass the arguments to the factory */
		RP p(f);                          /* create a resource pool */
						  /* Initialize a PrintFetcher for the specified number of columns */
		PrintFetcher fetch(parseString(argv[2]));
						  /* pass the SQL and the Fetcher to the SQLSelect class */	
		RP::Command::SQLSelect select(p, argv[3], fetch);
		fetch.bindout(select);            /* let the PrintFetcher bind the output parameters */
	
		p.execute(select);                /* execute the SQL */
	} catch (fatalmind::Exception e) {    /* if you provide missing/wrong data */
		std::cout << "An error happend: " << e.message() << std::endl;
	}
}

int parseString(char* param) {
	std::istringstream is(param);
	int rv;
	is >> rv;
	if (is.fail() || rv < 1) {
		throw fatalmind::Exception("Illegal argument (need numeric number of columns)");
	}
	return rv;
}
