#include "main.hh"
#include "config.h"

int fatalmind_trace_malloc = 0;

#ifdef HAVE_MALLINFO
#include <malloc.h>
#endif


// TODO, ifdef
#include <map>
/* Prototypes for our hooks.  */

#ifdef HAVE_MALLINFO
static void my_init_hook (void);
static void *my_malloc_hook (size_t, const void *);
static void my_free_hook (void*, const void *);

/* Override initializing hook from the C library. */

void (*__malloc_initialize_hook) (void) = my_init_hook;

void *(* old_malloc_hook) __MALLOC_PMT ((size_t __size, __const __malloc_ptr_t));
void (*old_free_hook) __MALLOC_PMT ((void *__ptr, __const __malloc_ptr_t));

typedef std::map<const void*, const void*>  malloctrace_t;
malloctrace_t malloctrace;

static void my_init_hook (void) {
	old_malloc_hook = __malloc_hook;
	old_free_hook = __free_hook;

	__malloc_hook = my_malloc_hook;
	__free_hook = my_free_hook;
}

static void * my_malloc_hook (size_t size, const void *caller) {
	void *result;

	/* Restore all old hooks */
	__malloc_hook = old_malloc_hook;
	__free_hook = old_free_hook;

	/* Call recursively */
	result = malloc (size);

	/* Save underlying hooks */
	old_malloc_hook = __malloc_hook;
	old_free_hook = __free_hook;

	/* printf might call malloc, so protect it too. */

	if (fatalmind_trace_malloc) {
		malloctrace[result] = caller;
	}

	/* Restore our own hooks */

	__malloc_hook = my_malloc_hook;
	__free_hook = my_free_hook;

	return result;
}

     

static void my_free_hook (void *ptr, const void *caller) {
       /* Restore all old hooks */
	__malloc_hook = old_malloc_hook;
	__free_hook = old_free_hook;

	/* Call recursively */
	free (ptr);

	/* Save underlying hooks */
	old_malloc_hook = __malloc_hook;

	old_free_hook = __free_hook;

	/* printf might call free, so protect it too. */

	if (fatalmind_trace_malloc) {
		malloctrace.erase(ptr);
	}

	/* Restore our own hooks */
	__malloc_hook = my_malloc_hook;
	__free_hook = my_free_hook;
}
#endif
// TODO
// TODO
// TODO
// TODO
// TODO

typedef unsigned int very_strange_fatalmind_typedef;

template<>
void _assertEqual<std::string>(
	  fatalmind::arg_traits<std::string>::const_param is
	, fatalmind::arg_traits<std::string>::const_param expect
	, const char* file
	, int line)
{
	ostringstream os;
	std::string::const_iterator eiis    = is.end();
	std::string::const_iterator eexpect = expect.end();
	std::string::const_iterator iis     = is.begin();
	std::string::const_iterator iexpect = expect.begin();
	for (; (iis != eiis) && (iexpect != eexpect); ++iis, ++iexpect) {
		if (*iis != *iexpect) {
			size_t pos = distance(is.begin(), iis);
			os << "assertEqual failed in " << file << ":"
				<< line << " (at position " << (pos) 
				<< " expected '" << *iexpect << "' (" 
					<< std::hex << std::setw(2) << (int(*iexpect)&0xFF)
				<< ") got '" << *iis << "' (" 
					<< std::hex << (int(*iis)&0xFF)
				<< ") actual result '" << is.substr(pos, min<size_t>(35, is.size()-pos)) << "'" << ((pos+35 < is.size())?"..." :"")<< ")";
			throw fatalmind::Exception(os.str());
		}
	}
	if (is.size() != expect.size()) {
		os << "assertEqual failed in " << file << ":"
			<< line << " (size is " << is.size() 
			<< " expected " << expect.size() << ")";
		throw fatalmind::Exception(os.str());
	}
}

void _assertEqual(const int is, const int expect, const char* file, int line) {
	_assertEqual<int>(is, expect, file, line);
}

void _assertEqual(const std::string& is, const std::string expect, const char* file, int line) {
	_assertEqual<std::string>(is, expect, file, line);
}

long getVirtualMemorySize();
void _assertNotGrown(fatalmind_test_size& sz, int iter, const char* file, int line) {
	return; // DISABLED FOR NOW, NOT YET STABLE
#ifdef notdefined
	if (iter >= 0) {
		if (iter == 0) {
			unsigned long virt = getVirtualMemorySize();
			malloctrace.clear();
			fatalmind_trace_malloc = 1;
			// first observed iteration
			sz.ru_idrss = virt;
			sz.icount   = iter;
		} else if (iter < 100 || (iter < 1000 && (iter % 10) == 0) || ((iter % 100) == 0)) {
			unsigned long virt = getVirtualMemorySize();
			sz.icount   = iter;
			if (sz.ru_idrss <  virt) {
				fatalmind_trace_malloc = 0;
				
				ostringstream os;
				os << "assertNotGrown failed in " << file << ":"
					<< line << " (memory usage grew by " 
					<< virt - sz.ru_idrss << " (from "
					<< sz.ru_idrss << " to " << virt 
					<< ") in iteration " << iter << ")";
				malloctrace_t::iterator end = malloctrace.end();
				for (malloctrace_t::iterator i = malloctrace.begin(); i != end; ++i) {
					os << "Ptr=" << i->first << " allocated from=" << i->second << ";";	
				}
				throw fatalmind::Exception(os.str());
			}
		}
	}
	sz.icount++;
#endif
}

void dumpLeakedMemory() {
/* DISABLED FOR NOW
	malloctrace_t::iterator end = malloctrace.end();
	for (malloctrace_t::iterator i = malloctrace.begin(); i != end; ++i) {
		std::cout << "Ptr=" << i->first << " allocated from=" << i->second << std::endl;;	
	}
	std::cout << distance(malloctrace.begin(), malloctrace.end()) << " objects leaked" << std::endl;
*/
}

// TODO, non portable :(
// getruusage returns always 0
long getVirtualMemorySize() {
	long rv = 0;
#ifdef HAVE_MALLINFO
	struct mallinfo m = mallinfo();
	rv = m.uordblks;
#else
#endif
	return rv;
}
